/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <ctype.h>
#include "ftp_main.h"

static struct portrangestruct *pasvrangeclient = (struct portrangestruct*) 0;
static struct portrangestruct *pasvrangeserver = (struct portrangestruct*) 0;
static void saypasv(struct messagebuffer*, int, char*, unsigned short int);


/* pasvserver() handles nearly the whole passive connection, i.e., it
 * requests a port on the server, binds to one on the proxy host and sends
 * the address & port to the client
 * 
 * Return values:
 * 	TRANSFERMODE_OK
 * 	TRANSFERMODE_FAIL
 * 	TRANSFERMODE_FATAL
 *
 * */

int pasvserver(struct proxyinfo* pi, struct messagebuffer* msgbuf,
					struct ftp_message *server_msg) {
	struct clientinfo* ci = pi->ci;
	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	int ret, pssock;
	char* brk;
	struct sockaddr_in pasvserv_sin;
	struct portrangestruct* prs;

	say(ss, "PASV\r\n");
	*server_msg = ftp_readall(ss);
	if (server_msg->error) {
		/* Treat the timeout with a special message, but default
		 * behaviour for the other cases */
		if (server_msg->error_type == ERROR_TIMEOUT) {
			jlog(5, "Timed out reading answer to the PASV command"
					" from fd %d", server_msg->from_fd);
			message_buffer_add(msgbuf, cs,
					500, "Timeout waiting for an answer "
						"for the PASV command\r\n");
			return TRANSFERMODE_FATAL;
		} else if (server_msg->error_type == ERROR_CONNECTIONCLOSED ) {
			message_buffer_add(msgbuf, cs,
					500, "Remote server closed the "
						"connection\r\n");
			return TRANSFERMODE_FATAL;
		} else {
			ftp_message_say_error(*server_msg, msgbuf, cs,
							__FILE__, __LINE__);
			return TRANSFERMODE_FAIL;
		}
	}

	/* 
	 * An example of the answer would be:
	 * "227 Entering Passive Mode (127,0,0,1,7,123)"
	 *
	 * Roxen Challenger at ftp.mutt.org sends this reply
	 * "227 Entering Passive Mode. 217,69,76,44,156,68"
	 */
	/* Look for the beginning of the IP */
	brk = (char*) strchr(server_msg->msg, '(');

	if (server_msg->code != 227) {
		/* invalid return string */
		char* msg_no_codes = strip_respcodes(server_msg->msg);
		jlog(2, "Strange answer after sending PASV: %s",
						server_msg->msg);
		message_buffer_add_nomalloc(msgbuf, cs,
					server_msg->code, msg_no_codes);
		return TRANSFERMODE_FAIL;
	}
	if (!brk) {
		/* don't give up, maybe it just didn't send a bracket */
		/* skip over the numerical code */
		brk = strchr(server_msg->msg, ' ');
		brk++;
		/* increase until we found a numeric */
		while (*brk && !isdigit((int)*brk)) {
			brk++;
		}
		/* now brk should point to the first numeric */
	} else {
		/* skip over the bracket sign */
		brk++;
	}

	/* call parsesock to parse the answer string and get the values into
	 * PASVSERVERSOCK */
	ret = parsesock(brk, &pasvserv_sin, PASSIVE);
	if (ret) {
		message_buffer_add(msgbuf, cs,
				500, "Could not parse a valid address "
				     "from the PASV response\r\n");
		jlog(2, "Could not parse a valid socket from the PASV "
		    " response: %s", server_msg->msg);
		return TRANSFERMODE_FAIL;
	}

	if (!pasvrangeserver) {
		const char* range = config_get_option("passiveportrangeserver");
		if (!range) {
			range = config_get_option("passiveportrange");
		}
		pasvrangeserver = config_parse_portranges(range);
	}
	if (!pasvrangeserver) {
		prs = portrangestruct_port2portrange(ci->data_connection.dataport);
	} else {
		prs = pasvrangeserver;
	}

	jlog(7, "Opening passive port %s:%d",
					inet_ntoa(pasvserv_sin.sin_addr),
					ntohs(pasvserv_sin.sin_port));
	/* open the port on the foreign machine specified by PASVSERVERSOCK */
	pssock = openportiaddr(pasvserv_sin.sin_addr.s_addr, /* dest ip   */
			/* dest port */
			ntohs(pasvserv_sin.sin_port),
			/* source ip */
			/* this variable is set after a successul login and
			 * respects the dataserveraddress option */
			ci->data_connection.data_addr_to_server,
			/* source ports */
			prs,
			pi);
	if (pssock < 0) {
		message_buffer_add(msgbuf, cs,
				500, "Could not connect to the specified "
				     "PASV host\r\n");
		jlog(3, "Could not connect to the specified PASV host: (%s)",
						server_msg->msg);
		return TRANSFERMODE_FAIL;
	}
	ci->data_connection.dataserversock = pssock;
	/* XXX free portrange? */

	return TRANSFERMODE_OK;
}


/* pasvclient() opens a port on the proxy machine and tells the addr and
 * port to the client
 *
 * Parameters: clntinfo: connection information
 *
 * Return values: -1 on error
 *                 0 on success
 *
 * Called by: handlecmds (if the client issued a PORT or PASV command)
 */


int pasvclient(struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;
	int cs = pi->si->connection.clientsocket;
	int pcsock;
	struct sockaddr_in pasvclientsin;
	const char* opt;
	struct ip_t display_ip;
	char* display_char;

	ci->data_connection.transfermode.clientmode = PASSIVE;

	if (!pasvrangeclient) {
		const char* range = config_get_option("passiveportrangeclient");
		if (! range) {
			range = config_get_option("passiveportrange");
		}
		pasvrangeclient = config_parse_portranges(range);
	}
	if (ci->data_connection.dataclientsock >= 0) {
		/* we are still listening on another socket. Close it */
		close(ci->data_connection.dataclientsock);
		ci->data_connection.dataclientsock = -1;
	}
	errno = 0;
	/* open a port on the interface the client is connected to and store
	 * the values in PASVCLIENTSIN */
	pcsock = openlocalport(&pasvclientsin,
			       ci->data_connection.data_addr_to_client,
			       pasvrangeclient,
			       pi);
	if (pcsock < 0) {
		/* XXX */
		/*say(cs, "500 Error binding to a port\r\n");*/
		jlog(2, "Could not bind (%s)", strerror(errno));
		return -1;
	}
	ci->data_connection.dataclientsock = pcsock;
	ci->data_connection.waitforconnect = &(ci->data_connection.dataclientsock);

	/* write the values to the client socket CS */
	if ((opt = config_get_option("displayclientaddress"))) {
		display_ip = ip_parse_from_char(opt);
		if (display_ip.error) {
			jlog(5, "Could not parse IP from option displayclientaddress: %s", opt);
		}
	} else if (pasvclientsin.sin_addr.s_addr == INADDR_ANY) {
		/* the user can specify 0.0.0.0 with the dataclientaddress
		 * parameter. In this case, we do not want to display
		 * 0,0,0,0,x,y in a response... */
		display_ip = ip_from_s_addr(si->connection.addr_to_client);
	} else {
		display_ip = ip_from_s_addr(pasvclientsin.sin_addr.s_addr);
	}
	display_char = ip_to_mchar(display_ip);
	saypasv(msgbuf, cs, display_char, htons(pasvclientsin.sin_port));
	free(display_char);

	return CMD_HANDLED;
}


static
void saypasv(struct messagebuffer* msgbuf, int fd, char* addr,
						unsigned short int port) {
	unsigned short int lo, hi;

	hi = port % 256;
	lo = (port - hi) / 256;

	replace_not_larger(addr, ".", ",");

	message_buffer_addf(msgbuf, fd,
		227, "Entering Passive Mode (%s,%d,%d)\r\n", addr, lo, hi);
}

void destroy_passive_portrange() {
	config_destroy_portrange(pasvrangeserver);
	config_destroy_portrange(pasvrangeclient);
	pasvrangeclient = (struct portrangestruct*) 0;
	pasvrangeserver = (struct portrangestruct*) 0;
}

