/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "../ftp_main.h"

static int forward_type(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf);
static int forward_port(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf);
static int forward_pasv(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf);
static int forward_sendcmd(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf);

struct config_directives_list nodatarelay_configuration_directives_array[] = {
	{"enabled",	TAG_ALL, "no", EM, WSP },
	{ (char*) 0,	0, (char*) 0, 0, 0 }
};

struct config_namespace_directives_list nodatarelay_configuration_directives = {
	"nodatarelay",		/* namespace name */
	nodatarelay_configuration_directives_array
};

struct config_option_values_list nodatarelay_configuration_options_array[] = {
	{"enabled",	{ TRUEFALSE, TERM } },
	{ TERM,         { TERM } }
};

struct config_namespace_options_list nodatarelay_configuration_options = {
	"nodatarelay",
	nodatarelay_configuration_options_array
};

int nodatarelay_add_configuration() {
	config_configuration_list_add(&nodatarelay_configuration_directives);
	config_options_list_add(&nodatarelay_configuration_options);
	return 0;
}

int nodatarelay_config_init() {
	config_register_hook(0, nodatarelay_add_configuration);
	return 0;
}

int nodatarelay_commandhandler_init() {
	if (config_get_bool("nodatarelay.enabled", 0) == 0) {
		return 0;
	}
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"RETR", "forwarded RETR", 1, 1,
			-1,
			forward_sendcmd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"STOR", "forwarded STOR", 1, 1,
			-1,
			forward_sendcmd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"APPE", "forwarded APPE", 1, 1,
			-1,
			forward_sendcmd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"LIST", "forwarded LIST", 0, 1,
			-1,
			forward_sendcmd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"NLST", "forwarded NLST", 0, 1,
			-1,
			forward_sendcmd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"TYPE", "forwarded TYPE", 1, 1,
			-1,
			forward_type,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PORT", "forwarded PORT", 1, 1,
			-1,
			forward_port,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASV", "forwarded PASV", 0, 0,
			-1,
			forward_pasv,
			(char*) 0, (cmd_hookfunc_t) 0);
	return 0;
}

int nodatarelay_init() {
	module_hook_register(MODULE_HOOK_PROXY_INIT_CONFIG,
			     nodatarelay_config_init,
			     "nodatarelay config init");

	module_hook_register(MODULE_HOOK_FTP_CMDHANDLER_INIT,
			     nodatarelay_commandhandler_init,
			     "nodatarelay commandhandler init");
	return 0;
}

static int forward_type(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 200);
}

static int forward_port(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 200);
}

static int forward_pasv(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 227);
}

static int forward_sendcmd(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	struct ftp_message msg;
	char* msg_no_codes;
	int code;
	int i;
	/* we can have
	 *
	 * 125
	 * 150 for initiation
	 *
	 * 425 for error (could not establish connection)
	 *
	 * 226 for completion
	 *
	 * 550 for error (not found)
	 * XXX abort ?
	 *
	 */

	/* At least for SITE we have more than one argument */
	say(ss, cmd);
	for (i = 0; args[i]; i++) {
		sayf(ss, " %s", args[i]);
	}
	say(ss, "\r\n");
	do {
		msg = ftp_readall(ss);
		if (msg.error) {
			ftp_message_say_error(msg, msgbuf, cs,
							__FILE__, __LINE__);
			return CMD_ABORT;
		}
		msg_no_codes = strip_respcodes(msg.msg);
		code = msg.code;
		switch(msg.code) {
		case 125:
		case 150:
		case 226:
			message_buffer_add_nomalloc(msgbuf, cs,
						msg.code, msg_no_codes);
			ftp_message_free(&msg);
			if (code == 125 || code == 150) {
				message_buffer_say_reset(msgbuf, cs, &code);
			}
			break;
		default:
			term_before_last_newline(msg_no_codes);
			message_buffer_addf(msgbuf, cs,
					msg.code, "%s\r\n", msg_no_codes);
			free(msg_no_codes);
			ftp_message_free(&msg);
			return CMD_ERROR;
		}
	} while (code != 226);
	return CMD_HANDLED;
}
